@@ -50,9 +50,9 @@ namespace Banshee.Dap.MassStorage
 {
     public class MassStorageSource : DapSource
     {
-        private Banshee.Collection.Gui.ArtworkManager artwork_manager 
+        private Banshee.Collection.Gui.ArtworkManager artwork_manager
             = ServiceManager.Get<Banshee.Collection.Gui.ArtworkManager> ();
-        
+
         private MassStorageDevice ms_device;
         private IVolume volume;
         private IUsbDevice usb_device;
@@ -60,15 +60,15 @@ namespace Banshee.Dap.MassStorage
         public override void DeviceInitialize (IDevice device)
         {
             base.DeviceInitialize (device);
-            
+
             volume = device as IVolume;
             if (volume == null || (usb_device = volume.ResolveRootUsbDevice ()) == null) {
                 throw new InvalidDeviceException ();
             }
-            
+
             ms_device = DeviceMapper.Map (this);
             try {
-                if (!ms_device.LoadDeviceConfiguration ()) {
+                if (ms_device.ShouldIgnoreDevice () || !ms_device.LoadDeviceConfiguration ()) {
                     ms_device = null;
                 }
             } catch {
@@ -78,7 +78,7 @@ namespace Banshee.Dap.MassStorage
             if (!HasMediaCapabilities && ms_device == null) {
                 throw new InvalidDeviceException ();
             }
-            
+
             // Ignore iPods, except ones with .is_audio_player files
             if (MediaCapabilities != null && MediaCapabilities.IsType ("ipod")) {
                 if (ms_device != null && ms_device.HasIsAudioPlayerFile) {
@@ -96,11 +96,11 @@ namespace Banshee.Dap.MassStorage
             mount_point = volume.MountPoint;
 
             Initialize ();
-            
+
             if (ms_device != null) {
                 ms_device.SourceInitialize ();
             }
-            
+
             AddDapProperties ();
 
             // TODO differentiate between Audio Players and normal Disks, and include the size, eg "2GB Audio Player"?
@@ -116,6 +116,13 @@ namespace Banshee.Dap.MassStorage
                 );
             }
 
+            if (VideoFolders.Length > 0 && !String.IsNullOrEmpty (VideoFolders[0])) {
+                AddDapProperty (String.Format (
+                    Catalog.GetPluralString ("Video Folder", "Video Folders", VideoFolders.Length), VideoFolders.Length),
+                    System.String.Join ("\n", VideoFolders)
+                );
+            }
+
             if (FolderDepth != -1) {
                 AddDapProperty (Catalog.GetString ("Required Folder Depth"), FolderDepth.ToString ());
             }
@@ -130,52 +137,56 @@ namespace Banshee.Dap.MassStorage
             }*/
         }
 
+        private System.Threading.ManualResetEvent import_reset_event;
         private DatabaseImportManager importer;
         // WARNING: This will be called from a thread!
         protected override void LoadFromDevice ()
         {
+            import_reset_event = new System.Threading.ManualResetEvent (false);
+
             importer = new DatabaseImportManager (this);
             importer.KeepUserJobHidden = true;
-            importer.Threaded = false; // We are already threaded
             importer.Finished += OnImportFinished;
 
             foreach (string audio_folder in BaseDirectories) {
                 importer.Enqueue (audio_folder);
             }
+
+            import_reset_event.WaitOne ();
         }
 
         private void OnImportFinished (object o, EventArgs args)
         {
             importer.Finished -= OnImportFinished;
 
-            if (!CanSyncPlaylists) {
-                return;
-            }
+            if (CanSyncPlaylists) {
+                var insert_cmd = new Hyena.Data.Sqlite.HyenaSqliteCommand (
+                    "INSERT INTO CorePlaylistEntries (PlaylistID, TrackID) VALUES (?, ?)");
+                int [] psources = new int [] {DbId};
+                foreach (string playlist_path in PlaylistFiles) {
+                    IPlaylistFormat loaded_playlist = PlaylistFileUtil.Load (playlist_path, new Uri (PlaylistsPath));
+                    if (loaded_playlist == null)
+                        continue;
 
-            Hyena.Data.Sqlite.HyenaSqliteCommand insert_cmd = new Hyena.Data.Sqlite.HyenaSqliteCommand (
-                "INSERT INTO CorePlaylistEntries (PlaylistID, TrackID) VALUES (?, ?)");
-            int [] psources = new int [] {DbId};
-            foreach (string playlist_path in PlaylistFiles) {
-                IPlaylistFormat loaded_playlist = PlaylistFileUtil.Load (playlist_path, new Uri (BaseDirectory));
-                if (loaded_playlist == null)
-                    continue;
-
-                PlaylistSource playlist = new PlaylistSource (System.IO.Path.GetFileNameWithoutExtension (playlist_path), this);
-                playlist.Save ();
-                //Hyena.Data.Sqlite.HyenaSqliteCommand.LogAll = true;
-                foreach (Dictionary<string, object> element in loaded_playlist.Elements) {
-                    string track_path = (element["uri"] as Uri).LocalPath;
-                    int track_id = DatabaseTrackInfo.GetTrackIdForUri (track_path, Paths.MakePathRelative (track_path, BaseDirectory), psources);
-                    if (track_id == 0) {
-                        Log.DebugFormat ("Failed to find track {0} in DAP library to load it into playlist {1}", track_path, playlist_path);
-                    } else {
-                        ServiceManager.DbConnection.Execute (insert_cmd, playlist.DbId, track_id);
+                    PlaylistSource playlist = new PlaylistSource (System.IO.Path.GetFileNameWithoutExtension (playlist_path), this);
+                    playlist.Save ();
+                    //Hyena.Data.Sqlite.HyenaSqliteCommand.LogAll = true;
+                    foreach (Dictionary<string, object> element in loaded_playlist.Elements) {
+                        string track_path = (element["uri"] as Uri).LocalPath;
+                        int track_id = DatabaseTrackInfo.GetTrackIdForUri (new SafeUri (track_path), psources);
+                        if (track_id == 0) {
+                            Log.DebugFormat ("Failed to find track {0} in DAP library to load it into playlist {1}", track_path, playlist_path);
+                        } else {
+                            ServiceManager.DbConnection.Execute (insert_cmd, playlist.DbId, track_id);
+                        }
                     }
+                    //Hyena.Data.Sqlite.HyenaSqliteCommand.LogAll = false;
+                    playlist.UpdateCounts ();
+                    AddChildSource (playlist);
                 }
-                //Hyena.Data.Sqlite.HyenaSqliteCommand.LogAll = false;
-                playlist.UpdateCounts ();
-                AddChildSource (playlist);
             }
+
+            import_reset_event.Set ();
         }
 
         public override void CopyTrackTo (DatabaseTrackInfo track, SafeUri uri, BatchUserJob job)
@@ -196,7 +207,7 @@ namespace Banshee.Dap.MassStorage
         public IVolume Volume {
             get { return volume; }
         }
-        
+
         public IUsbDevice UsbDevice {
             get { return usb_device; }
         }
@@ -264,7 +275,7 @@ namespace Banshee.Dap.MassStorage
                         }
                     }
                 }
-                
+
                 SupportsPlaylists &= CanSyncPlaylists;
                 return playlist_types;
             }
@@ -296,7 +307,11 @@ namespace Banshee.Dap.MassStorage
             }
 
             foreach (string file_name in PlaylistFiles) {
-                Banshee.IO.File.Delete (new SafeUri (file_name));
+                try {
+                    Banshee.IO.File.Delete (new SafeUri (file_name));
+                } catch (Exception e) {
+                    Log.Exception (e);
+                }
             }
 
             // Add playlists from Banshee to the device
@@ -304,19 +319,27 @@ namespace Banshee.Dap.MassStorage
             List<Source> children = new List<Source> (Children);
             foreach (Source child in children) {
                 PlaylistSource from = child as PlaylistSource;
-                if (from != null) {
+                string escaped_name = StringUtil.EscapeFilename (child.Name);
+                if (from != null && !String.IsNullOrEmpty (escaped_name)) {
                     from.Reload ();
                     if (playlist_format == null) {
                          playlist_format = Activator.CreateInstance (PlaylistTypes[0].Type) as PlaylistFormatBase;
                     }
 
                     SafeUri playlist_path = new SafeUri (System.IO.Path.Combine (
-                        PlaylistsPath, String.Format ("{0}.{1}", from.Name, PlaylistTypes[0].FileExtension)));
+                        PlaylistsPath, String.Format ("{0}.{1}", escaped_name, PlaylistTypes[0].FileExtension)));
 
-                    System.IO.Stream stream = Banshee.IO.File.OpenWrite (playlist_path, true);
-                    playlist_format.BaseUri = new Uri (BaseDirectory);
-                    playlist_format.Save (stream, from);
-                    stream.Close ();
+                    System.IO.Stream stream = null;
+                    try {
+                        stream = Banshee.IO.File.OpenWrite (playlist_path, true);
+                        playlist_format.BaseUri = new Uri (PlaylistsPath);
+
+                        playlist_format.Save (stream, from);
+                    } catch (Exception e) {
+                        Log.Exception (e);
+                    } finally {
+                        stream.Close ();
+                    }
                 }
             }
         }
@@ -328,11 +351,11 @@ namespace Banshee.Dap.MassStorage
             string [] names = ms_device != null ? ms_device.GetIconNames () : null;
             return names == null ? base.GetIconNames () : names;
         }
-        
+
         public override long BytesUsed {
             get { return BytesCapacity - volume.Available; }
         }
-        
+
         public override long BytesCapacity {
             get { return (long) volume.Capacity; }
         }
@@ -350,7 +373,7 @@ namespace Banshee.Dap.MassStorage
                     // According to the HAL spec, the first folder listed in the audio_folders property
                     // is the folder to write files to.
                     if (AudioFolders.Length > 0) {
-                        write_path = Banshee.Base.Paths.Combine (write_path, AudioFolders[0]);
+                        write_path = Hyena.Paths.Combine (write_path, AudioFolders[0]);
                     }
                 }
                 return write_path;
@@ -359,6 +382,25 @@ namespace Banshee.Dap.MassStorage
             set { write_path = value; }
         }
 
+        private string write_path_video = null;
+        public string WritePathVideo {
+            get {
+                if (write_path_video == null) {
+                    write_path_video = BaseDirectory;
+                    // Some Devices May Have a Separate Video Directory
+                    if (VideoFolders.Length > 0) {
+                        write_path_video = Hyena.Paths.Combine (write_path_video, VideoFolders[0]);
+                    } else if (AudioFolders.Length > 0) {
+                        write_path_video = Hyena.Paths.Combine (write_path_video, AudioFolders[0]);
+                        write_path_video = Hyena.Paths.Combine (write_path_video, "Videos");
+                    }
+                }
+                return write_path_video;
+            }
+
+            set { write_path_video = value; }
+        }
+
         private string [] audio_folders;
         protected string [] AudioFolders {
             get {
@@ -370,6 +412,17 @@ namespace Banshee.Dap.MassStorage
             set { audio_folders = value; }
         }
 
+        private string [] video_folders;
+        protected string [] VideoFolders {
+            get {
+                if (video_folders == null) {
+                    video_folders = HasMediaCapabilities ? MediaCapabilities.VideoFolders : new string[0];
+                }
+                return video_folders;
+            }
+            set { video_folders = value; }
+        }
+
         protected IEnumerable<string> BaseDirectories {
             get {
                 if (AudioFolders.Length == 0) {
@@ -386,7 +439,7 @@ namespace Banshee.Dap.MassStorage
         protected int FolderDepth {
             get {
                 if (folder_depth == -1) {
-                    folder_depth = HasMediaCapabilities ? MediaCapabilities.FolderDepth : 0;
+                    folder_depth = HasMediaCapabilities ? MediaCapabilities.FolderDepth : -1;
                 }
                 return folder_depth;
             }
@@ -440,33 +493,40 @@ namespace Banshee.Dap.MassStorage
                 Directory.Create (System.IO.Path.GetDirectoryName (new_uri.LocalPath));
                 File.Copy (fromUri, new_uri, false);
 
-                // TODO If write-to-file is NOT on, then write the metadata in db to the file on the DAP
-                // to ensure that when we load it next time, it's data will match what's in the database
-                // and the MetadataHash will actually match.  This isn't necessary on iPods or MTP since
-                // on them we store our db's metadata in their dbs.
-
                 DatabaseTrackInfo copied_track = new DatabaseTrackInfo (track);
                 copied_track.PrimarySource = this;
                 copied_track.Uri = new_uri;
+
+                // Write the metadata in db to the file on the DAP if it has changed since file was modified
+                // to ensure that when we load it next time, it's data will match what's in the database
+                // and the MetadataHash will actually match.  We do this by comparing the time
+                // stamps on files for last update of the db metadata vs the sync to file.
+                // The equals on the inequality below is necessary for podcasts who often have a sync and
+                // update time that are the same to the second, even though the album metadata has changed in the
+                // DB to the feedname instead of what is in the file.  It should be noted that writing the metadata
+                // is a small fraction of the total copy time anyway.
+
+                if (track.LastSyncedStamp >= Hyena.DateTimeUtil.ToDateTime (track.FileModifiedStamp)) {
+                    Log.DebugFormat ("Copying Metadata to File Since Sync time >= Updated Time");
+                    bool write_metadata = Metadata.SaveTrackMetadataService.WriteMetadataEnabled.Value;
+                    bool write_ratings_and_playcounts = Metadata.SaveTrackMetadataService.WriteRatingsAndPlayCountsEnabled.Value;
+                    Banshee.Streaming.StreamTagger.SaveToFile (copied_track, write_metadata, write_ratings_and_playcounts);
+                }
+
                 copied_track.Save (false);
             }
 
-            if (CoverArtSize > -1 && !String.IsNullOrEmpty (CoverArtFileType) && 
-                    !String.IsNullOrEmpty (CoverArtFileName) && FolderDepth > 0) {
+            if (CoverArtSize > -1 && !String.IsNullOrEmpty (CoverArtFileType) &&
+                    !String.IsNullOrEmpty (CoverArtFileName) && (FolderDepth == -1 || FolderDepth > 0)) {
                 SafeUri cover_uri = new SafeUri (System.IO.Path.Combine (System.IO.Path.GetDirectoryName (new_uri.LocalPath),
                                                                          CoverArtFileName));
-                string coverart_id;
-                if (track.HasAttribute (TrackMediaAttributes.Podcast)) {
-                    coverart_id = String.Format ("podcast-{0}", Banshee.Base.CoverArtSpec.EscapePart (track.AlbumTitle));
-                } else {
-                    coverart_id = track.ArtworkId;
-                }                
-                
-                if (!File.Exists (cover_uri) && CoverArtSpec.CoverExists (coverart_id)) {       
+                string coverart_id = track.ArtworkId;
+
+                if (!File.Exists (cover_uri) && CoverArtSpec.CoverExists (coverart_id)) {
                     Gdk.Pixbuf pic = null;
-                    
+
                     if (CoverArtSize == 0) {
-                        if (CoverArtFileType == "jpg" || CoverArtFileType == "jpeg") {                        
+                        if (CoverArtFileType == "jpg" || CoverArtFileType == "jpeg") {
                             SafeUri local_cover_uri = new SafeUri (Banshee.Base.CoverArtSpec.GetPath (coverart_id));
                             Banshee.IO.File.Copy (local_cover_uri, cover_uri, false);
                         } else {
@@ -477,13 +537,13 @@ namespace Banshee.Dap.MassStorage
                     }
 
                     if (pic != null) {
-                        try {                        
+                        try {
                             byte [] bytes = pic.SaveToBuffer (CoverArtFileType);
                             System.IO.Stream cover_art_file = File.OpenWrite (cover_uri, true);
                             cover_art_file.Write (bytes, 0, bytes.Length);
                             cover_art_file.Close ();
                         } catch (GLib.GException){
-                            Log.DebugFormat ("Could convert cover art to {0}, unsupported filetype?", CoverArtFileType);
+                            Log.DebugFormat ("Could not convert cover art to {0}, unsupported filetype?", CoverArtFileType);
                         } finally {
                             Banshee.Collection.Gui.ArtworkManager.DisposePixbuf (pic);
                         }
@@ -498,11 +558,11 @@ namespace Banshee.Dap.MassStorage
                 if (ms_device != null && !ms_device.DeleteTrackHook (track)) {
                     return false;
                 }
-            
+
                 string track_file = System.IO.Path.GetFileName (track.Uri.LocalPath);
                 string track_dir = System.IO.Path.GetDirectoryName (track.Uri.LocalPath);
                 int files = 0;
-                
+
                 // Count how many files remain in the track's directory,
                 // excluding self or cover art
                 foreach (string file in System.IO.Directory.GetFiles (track_dir)) {
@@ -511,18 +571,18 @@ namespace Banshee.Dap.MassStorage
                         files++;
                     }
                 }
-                
+
                 // If we are the last track, go ahead and delete the artwork
                 // to ensure that the directory tree can get trimmed away too
                 if (files == 0 && CoverArtFileName != null) {
                     System.IO.File.Delete (Paths.Combine (track_dir, CoverArtFileName));
                 }
-                
+
                 Banshee.IO.Utilities.DeleteFileTrimmingParentDirectories (track.Uri);
             } catch (System.IO.FileNotFoundException) {
             } catch (System.IO.DirectoryNotFoundException) {
             }
-            
+
             return true;
         }
 
@@ -537,7 +597,7 @@ namespace Banshee.Dap.MassStorage
                 volume.Eject ();
             }
         }
-        
+
         protected override bool CanHandleDeviceCommand (DeviceCommand command)
         {
             try {
@@ -551,51 +611,68 @@ namespace Banshee.Dap.MassStorage
         private string GetTrackPath (TrackInfo track, string ext)
         {
             string file_path = null;
-            if (ms_device == null || !ms_device.GetTrackPath (track, out file_path)) {
-                file_path = FileNamePattern.CreateFromTrackInfo (track);
+
+            if (track.HasAttribute (TrackMediaAttributes.Podcast)) {
+                string album = FileNamePattern.Escape (track.DisplayAlbumTitle);
+                string title = FileNamePattern.Escape (track.DisplayTrackTitle);
+                file_path = System.IO.Path.Combine ("Podcasts", album);
+                file_path = System.IO.Path.Combine (file_path, title);
+            } else if (track.HasAttribute (TrackMediaAttributes.VideoStream)) {
+                string album = FileNamePattern.Escape (track.DisplayAlbumTitle);
+                string title = FileNamePattern.Escape (track.DisplayTrackTitle);
+                file_path = System.IO.Path.Combine (album, title);
+            } else if (ms_device == null || !ms_device.GetTrackPath (track, out file_path)) {
+                // If the folder_depth property exists, we have to put the files in a hiearchy of
+                // the exact given depth (not including the mount point/audio_folder).
+                if (FolderDepth != -1) {
+                    int depth = FolderDepth;
+                    string album_artist = FileNamePattern.Escape (track.DisplayAlbumArtistName);
+                    string track_album  = FileNamePattern.Escape (track.DisplayAlbumTitle);
+                    string track_number = FileNamePattern.Escape (String.Format ("{0:00}", track.TrackNumber));
+                    string track_title  = FileNamePattern.Escape (track.DisplayTrackTitle);
+
+                    if (depth == 0) {
+                        // Artist - Album - 01 - Title
+                        string track_artist = FileNamePattern.Escape (track.DisplayArtistName);
+                        file_path = String.Format ("{0} - {1} - {2} - {3}", track_artist, track_album, track_number, track_title);
+                    } else if (depth == 1) {
+                        // Artist - Album/01 - Title
+                        file_path = String.Format ("{0} - {1}", album_artist, track_album);
+                        file_path = System.IO.Path.Combine (file_path, String.Format ("{0} - {1}", track_number, track_title));
+                    } else if (depth == 2) {
+                        // Artist/Album/01 - Title
+                        file_path = album_artist;
+                        file_path = System.IO.Path.Combine (file_path, track_album);
+                        file_path = System.IO.Path.Combine (file_path, String.Format ("{0} - {1}", track_number, track_title));
+                    } else {
+                        // If the *required* depth is more than 2..go nuts!
+                        for (int i = 0; i < depth - 2; i++) {
+                            if (i == 0) {
+                                file_path = album_artist.Substring (0, Math.Min (i+1, album_artist.Length)).Trim ();
+                            } else {
+                                file_path = System.IO.Path.Combine (file_path, album_artist.Substring (0, Math.Min (i+1, album_artist.Length)).Trim ());
+                            }
+
+                        }
+
+                        // Finally add on the Artist/Album/01 - Track
+                        file_path = System.IO.Path.Combine (file_path, album_artist);
+                        file_path = System.IO.Path.Combine (file_path, track_album);
+                        file_path = System.IO.Path.Combine (file_path, String.Format ("{0} - {1}", track_number, track_title));
+                    }
+                } else {
+                    file_path = MusicLibrarySource.MusicFileNamePattern.CreateFromTrackInfo (track);
+                }
             }
 
-            file_path = System.IO.Path.Combine (WritePath, file_path);
+            if (track.HasAttribute (TrackMediaAttributes.VideoStream)) {
+              file_path = System.IO.Path.Combine (WritePathVideo, file_path);
+            } else {
+              file_path = System.IO.Path.Combine (WritePath, file_path);
+            }
             file_path += ext;
 
             return file_path;
-            
-            /*string artist = FileNamePattern.Escape (track.ArtistName);
-            string album = FileNamePattern.Escape (track.AlbumTitle);
-            string number_title = FileNamePattern.Escape (track.TrackNumberTitle);
-
-            // If the folder_depth property exists, we have to put the files in a hiearchy of
-            // the exact given depth (not including the mount point/audio_folder).
-            if (FolderDepth != -1) {
-                int depth = FolderDepth;
-
-                if (depth == 0) {
-                    // Artist - Album - 01 - Title
-                    file_path = System.IO.Path.Combine (file_path, String.Format ("{0} - {1} - {2}", artist, album, number_title));
-                } else if (depth == 1) {
-                    // Artist - Album/01 - Title
-                    file_path = System.IO.Path.Combine (file_path, String.Format ("{0} - {1}", artist, album));
-                    file_path = System.IO.Path.Combine (file_path, number_title);
-                } else if (depth == 2) {
-                    // Artist/Album/01 - Title
-                    file_path = System.IO.Path.Combine (file_path, artist);
-                    file_path = System.IO.Path.Combine (file_path, album);
-                    file_path = System.IO.Path.Combine (file_path, number_title);
-                } else {
-                    // If the *required* depth is more than 2..go nuts!
-                    for (int i = 0; i < depth - 2; i++) {
-                        file_path = System.IO.Path.Combine (file_path, artist.Substring (0, Math.Min (i, artist.Length)).Trim ());
-                    }
-
-                    // Finally add on the Artist/Album/01 - Track
-                    file_path = System.IO.Path.Combine (file_path, artist);
-                    file_path = System.IO.Path.Combine (file_path, album);
-                    file_path = System.IO.Path.Combine (file_path, number_title);
-                }
-            } else {
-                file_path = System.IO.Path.Combine (file_path, FileNamePattern.CreateFromTrackInfo (track));
-            }
-            */
         }
     }
 }
